@@ -1,6 +1,5 @@
 import getpass
 import json
-import lzma
 import os
 import platform
 import re
@@ -14,15 +13,16 @@ from functools import wraps
 from hashlib import md5
 from io import BytesIO
 from pathlib import Path
-from typing import Any, Callable, Iterator, List, Optional, Set, Union
+from typing import Any, Callable, ContextManager, IO, Iterator, List, Optional, Set, Union, cast
 
 import requests
 import urllib3  # type: ignore
 
 from .exceptions import *
-from .instaloadercontext import InstaloaderContext
+from .instaloadercontext import InstaloaderContext, RateController
+from .nodeiterator import NodeIterator, resumable_iteration
 from .structures import (Hashtag, Highlight, JsonExportable, Post, PostLocation, Profile, Story, StoryItem,
-                         save_structure_to_file, load_structure_from_file)
+                         load_structure_from_file, save_structure_to_file)
 
 
 def get_default_session_filename(username: str) -> str:
@@ -153,8 +153,10 @@ class Instaloader:
        txt file.
     :param storyitem_metadata_txt_pattern: :option:`--storyitem-metadata-txt`, default is empty (=none)
     :param max_connection_attempts: :option:`--max-connection-attempts`
-    :param commit_mode: :option:`--commit-mode`
     :param request_timeout: :option:`--request-timeout`, set per-request timeout (seconds)
+    :param rate_controller: Generator for a :class:`RateController` to override rate controlling behavior
+    :param resume_prefix: :option:`--resume-prefix`, or None for :option:`--no-resume`.
+    :param check_resume_bbd: Whether to check the date of expiry of resume files and reject them if expired.
 
     .. attribute:: context
 
@@ -178,9 +180,12 @@ class Instaloader:
                  storyitem_metadata_txt_pattern: str = None,
                  max_connection_attempts: int = 3,
                  request_timeout: Optional[float] = None,
-                 commit_mode: bool = False):
+                 rate_controller: Optional[Callable[[InstaloaderContext], RateController]] = None,
+                 resume_prefix: Optional[str] = "iterator",
+                 check_resume_bbd: bool = True):
 
-        self.context = InstaloaderContext(sleep, quiet, user_agent, max_connection_attempts, request_timeout)
+        self.context = InstaloaderContext(sleep, quiet, user_agent, max_connection_attempts,
+                                          request_timeout, rate_controller)
 
         # configuration parameters
         self.dirname_pattern = dirname_pattern or "{target}"
@@ -196,12 +201,8 @@ class Instaloader:
             else post_metadata_txt_pattern
         self.storyitem_metadata_txt_pattern = '' if storyitem_metadata_txt_pattern is None \
             else storyitem_metadata_txt_pattern
-        self.commit_mode = commit_mode
-        if self.commit_mode and not self.save_metadata:
-            raise InvalidArgumentException("Commit mode requires JSON metadata to be saved.")
-
-        # Used to keep state in commit mode
-        self._committed = None  # type: Optional[bool]
+        self.resume_prefix = resume_prefix
+        self.check_resume_bbd = check_resume_bbd
 
     @contextmanager
     def anonymous_copy(self):
@@ -223,7 +224,8 @@ class Instaloader:
             storyitem_metadata_txt_pattern=self.storyitem_metadata_txt_pattern,
             max_connection_attempts=self.context.max_connection_attempts,
             request_timeout=self.context.request_timeout,
-            commit_mode=self.commit_mode)
+            resume_prefix=self.resume_prefix,
+            check_resume_bbd=self.check_resume_bbd)
         yield new_loader
         self.context.error_log.extend(new_loader.context.error_log)
         new_loader.context.error_log = []  # avoid double-printing of errors
@@ -249,15 +251,9 @@ class Instaloader:
         if filename_suffix is not None:
             filename += '_' + filename_suffix
         filename += '.' + file_extension
-        # A post is considered "commited" if the json file exists and is not malformed.
-        if self.commit_mode:
-            if self._committed and os.path.isfile(filename):
-                self.context.log(filename + ' exists', end=' ', flush=True)
-                return False
-        else:
-            if os.path.isfile(filename):
-                self.context.log(filename + ' exists', end=' ', flush=True)
-                return False
+        if os.path.isfile(filename):
+            self.context.log(filename + ' exists', end=' ', flush=True)
+            return False
         self.context.get_and_write_raw(url, filename)
         os.utime(filename, (datetime.now().timestamp(), mtime.timestamp()))
         return True
@@ -355,7 +351,8 @@ class Instaloader:
         except UnicodeEncodeError:
             self.context.log('txt', end=' ', flush=True)
         with open(filename, 'wb') as text_file:
-            shutil.copyfileobj(BytesIO(bcaption), text_file)
+            with BytesIO(bcaption) as bio:
+                shutil.copyfileobj(cast(IO, bio), text_file)
         os.utime(filename, (datetime.now().timestamp(), mtime.timestamp()))
 
     def save_location(self, filename: str, location: PostLocation, mtime: datetime) -> None:
@@ -365,10 +362,29 @@ class Instaloader:
                            "https://maps.google.com/maps?q={0},{1}&ll={0},{1}\n".format(location.lat,
                                                                                         location.lng))
         with open(filename, 'wb') as text_file:
-            shutil.copyfileobj(BytesIO(location_string.encode()), text_file)
+            with BytesIO(location_string.encode()) as bio:
+                shutil.copyfileobj(cast(IO, bio), text_file)
         os.utime(filename, (datetime.now().timestamp(), mtime.timestamp()))
         self.context.log('geo', end=' ', flush=True)
 
+    def format_filename_within_target_path(self,
+                                           target: Union[str, Path],
+                                           owner_profile: Optional[Profile],
+                                           identifier: str,
+                                           name_suffix: str,
+                                           extension: str):
+        """Returns a filename within the target path.
+
+        .. versionadded:: 4.5"""
+        if ((format_string_contains_key(self.dirname_pattern, 'profile') or
+             format_string_contains_key(self.dirname_pattern, 'target'))):
+            profile_str = owner_profile.username.lower() if owner_profile is not None else target
+            return os.path.join(self.dirname_pattern.format(profile=profile_str, target=target),
+                                '{0}_{1}.{2}'.format(identifier, name_suffix, extension))
+        else:
+            return os.path.join(self.dirname_pattern.format(),
+                                '{0}_{1}_{2}.{3}'.format(target, identifier, name_suffix, extension))
+
     @_retry_on_connection_error
     def download_title_pic(self, url: str, target: Union[str, Path], name_suffix: str, owner_profile: Optional[Profile],
                            _attempt: int = 1) -> None:
@@ -389,16 +405,7 @@ class Instaloader:
         else:
             pic_bytes = http_response.content
             pic_identifier = md5(pic_bytes).hexdigest()[:16]
-        pic_extension = 'jpg'
-        if ((format_string_contains_key(self.dirname_pattern, 'profile') or
-             format_string_contains_key(self.dirname_pattern, 'target'))):
-            profile_str = owner_profile.username.lower() if owner_profile is not None else target
-            filename = os.path.join(self.dirname_pattern.format(profile=profile_str,
-                                                                target=target),
-                                    '{0}_{1}.{2}'.format(pic_identifier, name_suffix, pic_extension))
-        else:
-            filename = os.path.join(self.dirname_pattern.format(),
-                                    '{0}_{1}_{2}.{3}'.format(target, pic_identifier, name_suffix, pic_extension))
+        filename = self.format_filename_within_target_path(target, owner_profile, pic_identifier, name_suffix, 'jpg')
         content_length = http_response.headers.get('Content-Length', None)
         if os.path.isfile(filename) and (not self.context.is_logged_in or
                                          (content_length is not None and
@@ -506,7 +513,6 @@ class Instaloader:
 
         # Download the image(s) / video thumbnail and videos within sidecars if desired
         downloaded = True
-        self._committed = self.check_if_committed(filename)
         if self.download_pictures:
             if post.typename == 'GraphSidecar':
                 edge_number = 1
@@ -719,59 +725,77 @@ class Instaloader:
                             fast_update: bool = False,
                             post_filter: Optional[Callable[[Post], bool]] = None,
                             max_count: Optional[int] = None,
-                            total_count: Optional[int] = None) -> None:
+                            total_count: Optional[int] = None,
+                            owner_profile: Optional[Profile] = None) -> None:
         """
         Download the Posts returned by given Post Iterator.
 
-        ..versionadded:: 4.4
+        .. versionadded:: 4.4
+
+        .. versionchanged:: 4.5
+           Transparently resume an aborted operation if `posts` is a :class:`NodeIterator`.
 
         :param posts: Post Iterator to loop through.
-        :param target: Target name
-        :param fast_update: :option:`--fast-update`
-        :param post_filter: :option:`--post-filter`
-        :param max_count: Maximum count of Posts to download (:option:`--count`)
-        :param total_count: Total number of posts returned by given iterator
+        :param target: Target name.
+        :param fast_update: :option:`--fast-update`.
+        :param post_filter: :option:`--post-filter`.
+        :param max_count: Maximum count of Posts to download (:option:`--count`).
+        :param total_count: Total number of posts returned by given iterator.
+        :param owner_profile: Associated profile, if any.
         """
-        for number, post in enumerate(posts):
-            if max_count is not None and number >= max_count:
-                break
-            if total_count is not None:
-                self.context.log("[{0:{w}d}/{1:{w}d}] ".format(number + 1, total_count,
-                                                               w=len(str(total_count))),
-                                 end="", flush=True)
-            else:
-                if max_count is not None:
-                    self.context.log("[{0:{w}d}/{1:{w}d}] ".format(number + 1, max_count,
-                                                                   w=len(str(max_count))),
+        displayed_count = (max_count if total_count is None or max_count is not None and max_count < total_count
+                           else total_count)
+        sanitized_target = target
+        if isinstance(target, str):
+            sanitized_target = _PostPathFormatter.sanitize_path(target)
+        with resumable_iteration(
+                context=self.context,
+                iterator=posts,
+                load=load_structure_from_file,
+                save=save_structure_to_file,
+                format_path=lambda magic: self.format_filename_within_target_path(
+                    sanitized_target, owner_profile, self.resume_prefix or '', magic, 'json.xz'
+                ),
+                check_bbd=self.check_resume_bbd,
+                enabled=self.resume_prefix is not None
+        ) as (is_resuming, start_index):
+            for number, post in enumerate(posts, start=start_index + 1):
+                if max_count is not None and number > max_count:
+                    break
+                if displayed_count is not None:
+                    self.context.log("[{0:{w}d}/{1:{w}d}] ".format(number, displayed_count,
+                                                                   w=len(str(displayed_count))),
                                      end="", flush=True)
                 else:
-                    self.context.log("[{:3d}] ".format(number + 1), end="", flush=True)
-            if post_filter is not None:
-                try:
-                    if not post_filter(post):
-                        self.context.log("{} skipped".format(post))
-                        continue
-                except (InstaloaderException, KeyError, TypeError) as err:
-                    self.context.error("{} skipped. Filter evaluation failed: {}".format(post, err))
-                    continue
-            with self.context.error_catcher("Download {} of {}".format(post, target)):
-                # The PostChangedException gets raised if the Post's id/shortcode changed while obtaining
-                # additional metadata. This is most likely the case if a HTTP redirect takes place while
-                # resolving the shortcode URL.
-                # The `post_changed` variable keeps the fast-update functionality alive: A Post which is
-                # obained after a redirect has probably already been downloaded as a previous Post of the
-                # same Profile.
-                # Observed in issue #225: https://github.com/instaloader/instaloader/issues/225
-                post_changed = False
-                while True:
+                    self.context.log("[{:3d}] ".format(number), end="", flush=True)
+                if post_filter is not None:
                     try:
-                        downloaded = self.download_post(post, target=target)
-                        break
-                    except PostChangedException:
-                        post_changed = True
+                        if not post_filter(post):
+                            self.context.log("{} skipped".format(post))
+                            continue
+                    except (InstaloaderException, KeyError, TypeError) as err:
+                        self.context.error("{} skipped. Filter evaluation failed: {}".format(post, err))
                         continue
-                if fast_update and not downloaded and not post_changed:
-                    break
+                with self.context.error_catcher("Download {} of {}".format(post, target)):
+                    # The PostChangedException gets raised if the Post's id/shortcode changed while obtaining
+                    # additional metadata. This is most likely the case if a HTTP redirect takes place while
+                    # resolving the shortcode URL.
+                    # The `post_changed` variable keeps the fast-update functionality alive: A Post which is
+                    # obained after a redirect has probably already been downloaded as a previous Post of the
+                    # same Profile.
+                    # Observed in issue #225: https://github.com/instaloader/instaloader/issues/225
+                    post_changed = False
+                    while True:
+                        try:
+                            downloaded = self.download_post(post, target=target)
+                            break
+                        except PostChangedException:
+                            post_changed = True
+                            continue
+                    if fast_update and not downloaded and not post_changed:
+                        # disengage fast_update for first post when resuming
+                        if not is_resuming or number > 0:
+                            break
 
     @_requires_login
     def get_feed_posts(self) -> Iterator[Post]:
@@ -831,8 +855,10 @@ class Instaloader:
         """
         self.context.log("Retrieving saved posts...")
         assert self.context.username is not None  # safe due to @_requires_login; required by typechecker
-        self.posts_download_loop(Profile.from_username(self.context, self.context.username).get_saved_posts(), ":saved",
-                                 fast_update, post_filter, max_count=max_count)
+        node_iterator = Profile.from_username(self.context, self.context.username).get_saved_posts()
+        self.posts_download_loop(node_iterator, ":saved",
+                                 fast_update, post_filter,
+                                 max_count=max_count, total_count=node_iterator.count)
 
     @_requires_login
     def get_location_posts(self, location: str) -> Iterator[Post]:
@@ -887,18 +913,20 @@ class Instaloader:
                                  max_count=max_count)
 
     @_requires_login
-    def get_explore_posts(self) -> Iterator[Post]:
+    def get_explore_posts(self) -> NodeIterator[Post]:
         """Get Posts which are worthy of exploring suggested by Instagram.
 
         :return: Iterator over Posts of the user's suggested posts.
+        :rtype: NodeIterator[Post]
         :raises LoginRequiredException: If called without being logged in.
         """
-        data = self.context.get_json('explore/', {})
-        yield from (Post(self.context, node)
-                    for node in self.context.graphql_node_list("df0dcc250c2b18d9fd27c5581ef33c7c",
-                                                               {}, 'https://www.instagram.com/explore/',
-                                                               lambda d: d['data']['user']['edge_web_discover_media'],
-                                                               data.get('rhx_gis')))
+        return NodeIterator(
+            self.context,
+            'df0dcc250c2b18d9fd27c5581ef33c7c',
+            lambda d: d['data']['user']['edge_web_discover_media'],
+            lambda n: Post(self.context, n),
+            query_referer='https://www.instagram.com/explore/',
+        )
 
     def get_hashtag_posts(self, hashtag: str) -> Iterator[Post]:
         """Get Posts associated with a #hashtag.
@@ -969,7 +997,7 @@ class Instaloader:
         .. versionadded:: 4.3"""
         self.context.log("Retrieving IGTV videos for profile {}.".format(profile.username))
         self.posts_download_loop(profile.get_igtv_posts(), profile.username, fast_update, post_filter,
-                                 total_count=profile.igtvcount)
+                                 total_count=profile.igtvcount, owner_profile=profile)
 
     def _get_id_filename(self, profile_name: str) -> str:
         if ((format_string_contains_key(self.dirname_pattern, 'profile') or
@@ -1079,11 +1107,11 @@ class Instaloader:
         def _error_raiser(_str):
             yield
 
-        # error_handler type is Callable[[Optional[str]], ContextManager[None]] (not supported with Python 3.5)
-        error_handler = _error_raiser if raise_errors else self.context.error_catcher
+        error_handler = cast(Callable[[Optional[str]], ContextManager[None]],
+                             _error_raiser if raise_errors else self.context.error_catcher)
 
         for profile in profiles:
-            with error_handler(profile.username):  # type: ignore
+            with error_handler(profile.username):
                 profile_name = profile.username
 
                 # Download profile picture
@@ -1124,7 +1152,7 @@ class Instaloader:
                 if posts:
                     self.context.log("Retrieving posts from profile {}.".format(profile_name))
                     self.posts_download_loop(profile.get_posts(), profile_name, fast_update, post_filter,
-                                             total_count=profile.mediacount)
+                                             total_count=profile.mediacount, owner_profile=profile)
 
         if stories and profiles:
             with self.context.error_catcher("Download stories"):
@@ -1204,26 +1232,7 @@ class Instaloader:
         # Iterate over pictures and download them
         self.context.log("Retrieving posts from profile {}.".format(profile_name))
         self.posts_download_loop(profile.get_posts(), profile_name, fast_update, post_filter,
-                                 total_count=profile.mediacount)
-
-    def check_if_committed(self, filename: str) -> bool:
-        """Checks to see if the current post has been committed.
-
-        A post is considered committed if its json metadata file exists and is not malformed.
-
-        .. versionadded:: 4.2
-        """
-        if os.path.isfile(filename + '.json.xz'):
-            filename += '.json.xz'
-        elif os.path.isfile(filename + '.json'):
-            filename += '.json'
-        else:
-            return False
-        try:
-            load_structure_from_file(self.context, filename)
-            return True
-        except (FileNotFoundError, lzma.LZMAError, json.decoder.JSONDecodeError):
-            return False
+                                 total_count=profile.mediacount, owner_profile=profile)
 
     def interactive_login(self, username: str) -> None:
         """Logs in and internally stores session, asking user for password interactively.
